#include <linux/module.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <asm/segment.h>
#include <asm/uaccess.h>
/* #include "/usr/src/linux-2.4.18-10/include/asm-i386/uaccess.h"*/
#include <asm/io.h>
#include <linux/delay.h>
/* #include "/usr/src/linux-2.4.18-10/include/asm-i386/delay.h"*/

/* PCI headers for kernel 2.2.xx */
#include <linux/config.h>
#include <linux/pci.h>  

static int dio_counter=0;

#include "dio.h"

/* define which parms can be altered */

#define is_hex(c)     (((c) >= '0' && (c) <= '9') || ((c)>='a' && (c)<='f') || ((c)>='A' && (c) <= 'F'))    
#define DIO_MOD_NAME	"dio"
static int dio_major=DIO_MAJOR;

MODULE_PARM(dio_major,"i");
MODULE_PARM(dio_minor,"i");
/*MODULE_AUTHOR("Matt Clark, Aug 2001");*/

static int dio_ioctl(struct inode *, struct file *, unsigned int , unsigned long);
static ssize_t dio_read(struct file *, char *, size_t, loff_t *);
static ssize_t dio_write(struct file *, const char *, size_t, loff_t *);
static int dio_open(struct inode *, struct file *);
static int dio_release(struct inode *, struct file *);
inline static int dio_handle_buffer(dio_data_t * ,int);
inline static dio_delay(void);
static float dio_adconv(int);  
static BOOL dio_text;

struct file_operations dio_fops = {
	read:		dio_read,
	write:		dio_write,
	ioctl:		dio_ioctl,
	open:		dio_open,
	release:	dio_release,
//	mmap:		dio_mmap,
	};

int init_module(void) {
int i,res;
volatile u32	temp;
struct pci_dev*	dev=NULL;

/* refuse to load if PCI is not native to the machine */
#ifndef CONFIG_PCI
	printk("pcitest:pci bus is not supported on this computer\n");
	return -ENODEV;
#endif   
/* refuse to run if the pci libraries are not supported in this kernel */
	if(!pcibios_present()){
		printk("dio: pci bus is present BUT pci support is not enabled in kernel\n");
		return -ENODEV;
		}										
/* look for a DIO card */
	if(!(dev=pci_find_device(DIO_VENDOR,DIO_DEVICE,dev))){
		printk("dio: no dio pci card found in this machine\n");
		return -ENODEV;
		}
/* next error should <never> occur (here for coding erros only)*/
	if(dev==NULL){
		printk("dio: card found but dev==NULL\n");
		return -ENODEV;
		}
/* now we have the device and a pointer to its device structure
 * extract the numbers we need using the dev structure rather than
 * the configuration registers (change since 2.1.xx matt oct 2001)
 * in case the kernel has rewritten the locations 
 */
/* now I know that base_address[0] is config
 *		 base_address[1] is the FPGA fuse map and private device code
 *		 base_address[2] is the I/O ports (as pci memory)  
 * I will assume that the base address is static once I get it 
 */
/* kernel version sepcific code here */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0)
	/* while we here bomb out at older kernels */
	printk("dio:kernels earlier than 2.2.xx not supported \n");
	return -ENODEV;
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
/* under kernel 2.2.xx base_address[] is a member of dev */
/* use ioremap to get the memory */
#define GET_BASE	dio_base=(unsigned int *)ioremap(dev->base_address[2]&PCI_BASE_ADDRESS_MEM_MASK,DIO_IO_REGION_SIZE);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
/* use  2.4 interface */
#define GET_BASE	dio_base=ioremap(pci_resource_start(dev,2)&PCI_BASE_ADDRESS_MEM_MASK,DIO_IO_REGION_SIZE);
#endif
	GET_BASE
	if(dio_base==0){
		printk("dio: init: could not ioremap pci address space\n");
		return -ENODEV;
		}
	dio_cntl=(unsigned int *)((void *)dio_base+DIO_CNTL);
	PRINT_DEBUG("dio: init: DIO at %x and CNTL at %x\n",(int) dio_base,(int) dio_cntl);
	

/* configure global variables - note these should be in dio_open
 * and part of dio_data struct for multiple opens / cards
 * and should be protected for race conditions and reentrancy 
 * under multiple open /card / processor configs
 * should be SMP safe under single access locking matt 2001
 */
	dio_text=FALSE; 
	dio_delay_time=0;
/* now configure the control word */
	temp=readl(dio_cntl);
	PRINT_DEBUG("dio:init:cntl=%x\n",temp);
	writel(0,dio_cntl);
	dio_dir=DIO_ALL_READ&DIO_CNTL_PORT_MASK;
	temp=readl(dio_cntl);
	PRINT_DEBUG("dio:init:cntl=%x\n",temp);
	writel((dio_dir|dio_latch)&DIO_CNTL_PORT_MASK,dio_cntl);
	temp=readl(dio_cntl);
	PRINT_DEBUG("dio:init:cntl=%x\n",temp);

/* now register the char device - never sure what order is best */
	res=register_chrdev(dio_major,DIO_MOD_NAME,&dio_fops);
	if (res<0) return(res);
	dio_major=res;
	
	printk("dio dev: init_module: Matt Clark aug 2001- device %d\n",dio_major);
	return 0;
	}     

	




static ssize_t dio_read(struct file * file, char * buf, size_t count, loff_t *ppos){
int length,buff_size,i;
dio_data_t *data;

/* check to see if any reads are enabled */
	if(dio_n_read==0) return 0;
/* check to see if private data is available (should always be but I don't want to oops)*/
	if(!(file->private_data)) return 0;
	data=(dio_data_t *)file->private_data;

	if(dio_text==TRUE){
		/* handle eof in text interface */
		if(data->eof_next==TRUE) return 0;
		/* we are opened as text- truncate the count and return
		 * a text string- the rest is handled in open */
		length=dio_n_read*sizeof(u32);
		if(length>count) return 0;
		count=length;
		}

/* quantise count to be u32 size maximum / len is in 32 bit words here*/
/* quantise cout to sizeof(u32)*n ports */
	length=count/sizeof(u32);
	if((buff_size=dio_handle_buffer(data,length))<0) return buff_size;
	/* clip the length to buffer length units */
	if(buff_size<length)length=buff_size;
/*quantise the size according to the number of reads */
	length=dio_n_read*(int)(length/dio_n_read);
/* printk("About to read %d u32s with a delay of %dus\n",length,dio_delay_time);*/
	for(i=0;i<length;){
		if(DIO_PORT0&dio_ports_read)data->buffer[i++]=readl(dio_base+DIO_IO_PORT_0);
		if(DIO_PORT1&dio_ports_read)data->buffer[i++]=readl(dio_base+DIO_IO_PORT_1);
		if(DIO_PORT2&dio_ports_read)data->buffer[i++]=readl(dio_base+DIO_IO_PORT_2);
		dio_delay();
		}	
/*	printk("dio:read dio_dir=%x, dio_ports_read=%x\n",dio_dir,dio_ports_read);*/
/*	printk("dio:read: got %x\n",data->buffer[0]);*/
/* convert len to bytes */
	if(dio_text==FALSE){
		length*=sizeof(u32);
		copy_to_user(buf,data->buffer,length);
		return length;
		}
	else{
		char text_buffer[DIO_TEXT_BUFF_SIZE],*pos;
		for(i=0,length=0;i<dio_n_read;i++){
			length+=sprintf(text_buffer+length,"%x ",data->buffer[i]);
			}
		text_buffer[length]='\n';
		text_buffer[++length]='\0';
		copy_to_user(buf,text_buffer,length);
		/* make sure eof is issued next time */
		data->eof_next=TRUE;
		return length;
		}
	}

static ssize_t dio_write(struct file * file,const  char * buf, size_t count, loff_t *ppos){
int length,buff_size,err,i,trans;
dio_data_t *data;


/* check to see if any writes are enabled */
	if(dio_n_write==0) return -EINVAL;
/* check to see if private data is available (should always be but I don't want to oops)*/
	if(!(file->private_data)) return -EFAULT;
	data=(dio_data_t *)file->private_data;
	
	if(dio_text==TRUE){
		char text_buffer[DIO_TEXT_BUFF_SIZE],*pos;
		
		/* as we are writing I have to interpret the input number as a string 
		 * not trivial as sscanf is not defined in the kernel */
		PRINT_DEBUG("dio:write:getting buffer\n");			
		memset(text_buffer,'\0',DIO_TEXT_BUFF_SIZE);
		copy_from_user(text_buffer,buf,count>DIO_TEXT_BUFF_SIZE-1?DIO_TEXT_BUFF_SIZE-1:count);
		PRINT_DEBUG("dio:interping buffer %s\n",text_buffer);
		for(i=0,pos=text_buffer;i<dio_n_write;i++){
			/* strip white (and non hex) space manually*/
			while((!is_hex(*pos))&((*pos!='\0')&((*pos)!='\n')))pos++;
			if((*pos)=='\0')break;
			if((*pos)=='\n')break;
			PRINT_DEBUG("dio:write:text reading on %s\n",pos);
			data->buffer[i]=simple_strtoul(pos,&pos,16);
			PRINT_DEBUG("dio:write:text now on %s\n",pos);
			}	
		if(i!=dio_n_write){
			printk("dio:write:text: badly formed text input\n");
			return -EINVAL;
			}
		if(i>=DIO_TEXT_BUFF_SIZE-1){
			printk("dio:write:warning: overflowed text buffer\n");
			return -EINVAL;
			}
		PRINT_DEBUG("dio:write:text got %s",text_buffer);
		for(i=0;i<dio_n_write;i++){
			PRINT_DEBUG("dio:write:text:writing %x\n",data->buffer[i]);
			}
		length=dio_n_write;
		/* lie to the functions- pretend we swallowed it all */
		trans=count;
		}
	else{
	/* quantise count to be u32 size maximum / len is in 32 bit words here*/
	/* quantise cout to sizeof(u32)*n ports */
		length=count/sizeof(u32);
		if((buff_size=dio_handle_buffer(data,length))<0) return buff_size;
	/* clip length if it is > buff_size*/
		if(length>buff_size)length=buff_size;
	/* quantise according to the number of port reads */
		length=dio_n_write*(int)(length/dio_n_write);	
	/* grab the data */
		copy_from_user(data->buffer,buf,length*sizeof(u32));
		PRINT_DEBUG("dio:bin:write: got %x\n",data->buffer[0]);
		PRINT_DEBUG("dio:bin:read dio_dir=%x, dio_ports_write=%x\n",dio_dir,dio_ports_write);
		trans=length*sizeof(u32);
		}
	/* write it to the ports */
	for(i=0;i<length;){
		if(DIO_PORT0&dio_ports_write)writel(data->buffer[i++],dio_base+DIO_IO_PORT_0);
		if(DIO_PORT1&dio_ports_write)writel(data->buffer[i++],dio_base+DIO_IO_PORT_1);
		if(DIO_PORT2&dio_ports_write)writel(data->buffer[i++],dio_base+DIO_IO_PORT_2);
		dio_delay();
		}	

	return trans;
	}



/* OPEN AND RELEASE */
/* we could use a static data structure instead of private data 
 * because we enforce global locking
 * however I prefer private data as it 
 * makes life easier in the lng run
 */
static int dio_open(struct inode *inode, struct file *file){ 
dio_data_t *data;
int err,minor;
u32	what=0;
/* simply enforce single open access */
/* global locking here prevents ioctl */
	//if(dio_locked==TRUE) return -EBUSY;
	dio_locked=TRUE;

	if((file->private_data=kmalloc(sizeof(dio_data_t),GFP_KERNEL))==NULL)
		return -1;
	data=(dio_data_t *)file->private_data;
	data->buffer=NULL;
	data->buff_size=0;
	if((err=dio_handle_buffer(data,DIO_BUFFER_SIZE))<0) return err;
	minor=MINOR(inode->i_rdev);
	if(minor==MINOR_BIN)dio_text=FALSE;
	else{
		/* here we are opened as text so use the minor number to
		 * get the port numbers and then set the read / write 
		 * permissions accordingly */
		/* do it only on a port wide basis- clunk I know but 
		 * convenient */
		dio_text=TRUE;
		if(minor>MINOR_TEXT){
			printk("dio:open:text:minor out of range\n");
			return -ENODEV;
			}
		if(minor&1)what|=DIO_PORT0;	
		if(minor&2)what|=DIO_PORT1;	
		if(minor&4)what|=DIO_PORT2;	
		PRINT_DEBUG("dio: opened as text\n");
		/* now use this to set dio_dir */
		if((file->f_flags & O_ACCMODE)==O_WRONLY){
			dio_dir|=what;
			dio_ports_write=what;
			dio_n_write=0;
			if(dio_ports_write&DIO_PORT0)dio_n_write++;
			if(dio_ports_write&DIO_PORT1)dio_n_write++;
			if(dio_ports_write&DIO_PORT2)dio_n_write++;   
			PRINT_DEBUG("dio:open:text:dio_dir=%x,dio_ports_write=%x,dio_n_write=%d\n",(int)dio_dir,(int)dio_ports_write,dio_n_write);
			}
		else{
			dio_dir&=((~what)&DIO_DIR_MASK);
			dio_ports_read=what;
			dio_n_read=0;
			if(dio_ports_read&DIO_PORT0)dio_n_read++;
			if(dio_ports_read&DIO_PORT1)dio_n_read++;
			if(dio_ports_read&DIO_PORT2)dio_n_read++;
			PRINT_DEBUG("dio:open:text:dio_dir=%x,dio_ports_read=%x,dio_n_read=%d\n",(int)dio_dir,(int)dio_ports_read,dio_n_read);
			}
		/* set the i/o port directions */
		writel((dio_dir|dio_latch)&DIO_CNTL_PORT_MASK,dio_cntl); 	
		/* set to read and then end!*/
		data->eof_next=FALSE;
		}
		
	MOD_INC_USE_COUNT;
	PRINT_DEBUG("dio:open:done\n");
	return 0;
	}

inline static int dio_handle_buffer(dio_data_t * data,int size){
	/* buffer is big enought */
	if(size<data->buff_size)return data->buff_size;
	
	/* buffer has reached its limit */
	if(data->buff_size==DIO_BUFF_LIMIT)return DIO_BUFF_LIMIT;
	
	/* asked for over the limit but the limit has not been reached yet */
	if(size>DIO_BUFF_LIMIT)size=DIO_BUFF_LIMIT;
	
	/* check to see if there was a buffer */
	if(data->buffer!=NULL) kfree(data->buffer);

	data->buffer=kmalloc(sizeof(u32)*size,GFP_KERNEL);
	if(data->buffer==NULL) return -ENOMEM;
	data->buff_size=size;
	PRINT_DEBUG("dio:handle_buffer:buffer is %d u32s\n",size);
	return size;
	}

static int dio_release(struct inode *inode, struct file *file) { 
dio_data_t *data;
	MOD_DEC_USE_COUNT;	
/* release the device */
	dio_locked=FALSE;

	if(file->private_data){
		data=(dio_data_t *)file->private_data;
		if(data->buffer!=NULL){
			kfree(data->buffer);
			data->buffer=NULL;
			data->buff_size=0;
			}
		kfree(file->private_data);
		file->private_data=NULL;
		}
	return 0;
	}	

static int dio_ioctl(struct inode * inode,struct file *file, unsigned int cmd, unsigned long arg){
	switch(cmd){
		case DIO_HARDRESET:
			/* resets the usage count to 1
			   for use after a module crash
			   make sure nothing else has opened
			   the module */
			printk("dio: ioctl: HARDRESET: attempting mod_use=1 ");
			/* pull it to 0 */
			while(MOD_IN_USE)MOD_DEC_USE_COUNT;
			MOD_INC_USE_COUNT;
			/* set locked to FALSE */
			dio_locked=FALSE;
			/* inc once for this open */
			printk("dio: Done\n");
			break;
		case DIO_LATCH:
			dio_latch=DIO_CNTL_LATCH_MASK;
			writel((dio_dir|dio_latch)&DIO_CNTL_PORT_MASK,dio_cntl);
			break;
		case DIO_UNLATCH:
			dio_latch=0;
			writel((dio_dir|dio_latch)&DIO_CNTL_PORT_MASK,dio_cntl);
			break;
		case DIO_DIR:
			dio_dir=arg;
			writel((dio_dir|dio_latch)&DIO_CNTL_PORT_MASK,dio_cntl);
			break;
		case DIO_READ_SEQ:
			dio_ports_read=(int)arg;
			dio_n_read=0;
			if(dio_ports_read&DIO_PORT0)dio_n_read++;
			if(dio_ports_read&DIO_PORT1)dio_n_read++;
			if(dio_ports_read&DIO_PORT2)dio_n_read++;
			break;
		case DIO_WRITE_SEQ:
			dio_ports_write=arg;	
			dio_n_write=0;
			if(dio_ports_write&DIO_PORT0)dio_n_write++;
			if(dio_ports_write&DIO_PORT1)dio_n_write++;
			if(dio_ports_write&DIO_PORT2)dio_n_write++;
		case DIO_PARALLEL:
			dio_par=TRUE;
			break;
		case DIO_SEQUENTIAL:
			dio_par=FALSE;
			break;
		case DIO_SET_DELAY:
#ifdef DIO_DELAY
			dio_delay_time=(int)arg;
			PRINT_DEBUG("dio:ioctl: %d dio_delay_time %d\n",dio_counter++,dio_delay_time);
#endif
#ifndef DIO_DELAY
		
#endif

			break;
		default:
			 PRINT_DEBUG("dio: ioctl: Unknown ioctl cmd %x\n",(int)cmd);
			}
	return 0;
	}  



void cleanup_module(void){
int res;
volatile u32 temp;
	printk("dio dev: cleanup_module: Matt Clark aug 2001- device %d\n",dio_major);
	if(dio_base!=NULL){
		iounmap(dio_base);
		dio_base=NULL;
		dio_cntl=NULL;
		}
	res=unregister_chrdev(dio_major,DIO_MOD_NAME);
	}    
inline static dio_delay(){
#ifdef DIO_DELAY
int jif;
	if(dio_delay_time){
		if(dio_delay_time<DIO_USECS_PER_HZ){
		/* busy wait for ~ the right time*/
			udelay(dio_delay_time);
			}
		else{
		/* sleep for ~ the right time */
			jif=(dio_delay_time*HZ)/1000000;
			if(jif<1)jif=1;
			current->state=TASK_INTERRUPTIBLE;
			schedule_timeout(jif);
			}
		}
#endif
	}
/* check out this from "/usr/src/linux-2.2.17/drivers/sound/cs4281.c" */
/*static void delayus(u32 delay)
{
        u32 j;
        if(delay > 9999)
        {
                j = (delay * HZ)/1000000;  
                if(j<1) j=1;              
                current->state = TASK_UNINTERRUPTIBLE;
                schedule_timeout(j);
        }
        else
                udelay(delay);
        return;
}  */ 
