/* device driver for dio dio card */
/* these defs are available to user programs */

/* ioctl numbers - arbitrary number here
 * consult Documentation/magic-number.txt
 * if you want to make a more informed choice !
 */
#include <linux/ioctl.h>
#define MAGIC		104

/* make it possible to remove the module if crashed */
#define DIO_HARDRESET		_IO(MAGIC,0)
#define DIO_LATCH		_IO(MAGIC,1)
#define DIO_UNLATCH		_IO(MAGIC,2)
#define DIO_DIR			_IOW(MAGIC,3,0)
#define DIO_READ_SEQ		_IOW(MAGIC,4,0)
#define DIO_WRITE_SEQ		_IOW(MAGIC,5,0)
#define DIO_PARALLEL		_IO(MAGIC,6)
#define DIO_SEQUENTIAL		_IO(MAGIC,7)
#define DIO_SET_DELAY		_IOW(MAGIC,8,0)
/*notice that size of the transfers is 0, I am using
 *called by value which is not convention
 */
#define DIO_ALL_READ		0x00000000ul
#define DIO_ALL_WRITE		0x00000FFFul
#define DIO_PORT0		0x0000000Ful
#define DIO_PORT1		0x000000F0ul
#define DIO_PORT2		0x00000F00ul
#define DIO_ALL_PORTS		DIO_PORT0|DIO_PORT1|DIO_PORT2


/* these are not exported to user progs */
#ifdef MODULE
#ifndef DIO_MAJOR
#define DIO_MAJOR 0
#endif
/* Minor dev numbers, 0->binary access 1,2,4 ports 0,1,2 1+2+4=allports */
#define MINOR_BIN	0
#define MINOR_TEXT	1+2+4 /* just use as a mask */


#ifdef DIO_DEBUG
#define PRINT_DEBUG(a,args...)	printk(a,## args);
#else 
#define PRINT_DEBUG(a,args...)
#endif


#define DIO_MOD_NAME		"dio"

/* following are the factory defaults for the DIO */
/* They are all acceptable for webface- matt aug 2001*/
#define DIO_VENDOR	0x1230
#define DIO_DEVICE	0x0001
#define BOOL	char
#define TRUE	1
#define FALSE 	0
#define DIO_READ 	1
#define	DIO_WRITE	0

/* a fixed text buffer > sizeof("12345678 12345678 12345678"); */

#define DIO_TEXT_BUFF_SIZE	100
/* I implement a growable buffer to accomodate any transfer size
 * It starts 512 bytes big (size of a sector read)
 */
#define DIO_BUFFER_SIZE		512/sizeof(u32)
/* enforce a 16k buffer limit */
#define DIO_BUFF_LIMIT		PAGE_SIZE*4/sizeof(u32)

#define DIO_IO_REGION_SIZE	64*1024 /*64k space */
#define DIO_WORD_SIZE		4
#define DIO_CNTL		0xc000
#define DIO_IO_PORT_0		0x8000
#define DIO_IO_PORT_1		0x8010
#define DIO_IO_PORT_2		0x8020
#define DIO_NO_DIO_PORTS	3	
#define DIO_NO_DIO_BYTES	DIO_NO_DIO_PORTS*DIO_WORD_SIZE
/* first 12 out of 32 bits =FFF bits 1-12 r/w mask, 13 latch, 14 reserved */
/* in the 1st 12 bits a 1 indicates output and 0 input */
#define DIO_CNTL_PORT_MASK	0x00008FFF
#define DIO_CNTL_LATCH_MASK	0x00008000
#define DIO_CNTRL_TEST_MASK	0x00004000
#define DIO_DIR_MASK		0x00000FFF
static int dio_ports_read	=	DIO_PORT0|DIO_PORT1|DIO_PORT2;
static int dio_ports_write	=	DIO_PORT0|DIO_PORT1|DIO_PORT2;

static u32 dio_dir;
static u32 dio_latch=FALSE;

/* timing */
/* busy wait for short delays, sleepon for long ones*/
/* DIO_USECS_PER_HZ determines the cross over */
/* if dio_par==TRUE, delay once per register group
 * else once per read (if you read on port it makes 
 * no difference 
 */
#define DIO_USECS_PER_HZ	1000000/HZ

/* driver globals */
static BOOL dio_locked=FALSE;
static BOOL dio_par=TRUE;
static dio_delay_time=0;  
static void * dio_base=NULL;
static void * dio_cntl=NULL; 
static int dio_n_read=0;
static int dio_n_write=0;


/* private data struct */
typedef struct dio_data{
	u32 * buffer;
	int buff_size;
	BOOL eof_next;
	} dio_data_t;


#endif
